import { ARRAY } from '../support/consts';
import { isMap, isObject, isPrimitive, isSet } from '../support/util';
import { attachMeta } from './meta';
export function deepCopy(obj) {
    const innerDeep = (obj) => {
        if (isPrimitive(obj)) {
            return obj;
        }
        // TODO: 或许这里通过 metaVer 能够解决多引用问题 ( see 3.4.2 )
        // TODO: 传入参数 reuse，确定是否走逻辑查看 meta.modified，并复用 meta.copy or meta.self
        let newNode = obj;
        if (Array.isArray(obj)) {
            newNode = obj.slice();
            newNode.forEach((item, idx) => {
                newNode[idx] = innerDeep(item);
            });
        }
        if (isSet(obj)) {
            const tmpArr = Array.from(obj);
            tmpArr.forEach((item, idx) => {
                tmpArr[idx] = innerDeep(item);
            });
            newNode = new Set(tmpArr);
        }
        if (isMap(obj)) {
            newNode = new Map(obj);
            newNode.forEach((value, key) => {
                newNode.set(key, innerDeep(value));
            });
        }
        if (isObject(obj)) {
            newNode = {};
            Object.keys(obj).forEach((key) => {
                newNode[key] = innerDeep(obj[key]);
            });
        }
        return newNode;
    };
    return innerDeep(obj);
}
/**
 * 尝试生成copy
 * @param val
 * @returns
 */
export function tryMakeCopy(val, options) {
    const { parentType, fastModeRange } = options;
    if (Array.isArray(val)) {
        return { copy: val.slice(), fast: false };
    }
    const fast = (fastModeRange === 'array' && parentType === ARRAY) || fastModeRange === 'all';
    let copy = val;
    if (val && isObject(val)) {
        copy = Object.assign({}, val);
    }
    if (isMap(val)) {
        copy = new Map(val);
    }
    if (isSet(val)) {
        copy = new Set(val);
    }
    return { copy, fast };
}
// 调用处已保证 meta 不为空
export function makeCopyWithMeta(ori, meta, options) {
    const { apiCtx, immutBase } = options;
    // LABEL: FIX IMMUT MEM LEAK
    if (immutBase) {
        return { copy: ori, fast: false };
    }
    const { copy, fast } = tryMakeCopy(ori, options);
    attachMeta(copy, meta, { apiCtx, fast });
    return { copy, fast };
}
